"""
This module provides functions for platform integration. Most of the functionality here
could also be achieved with psutils but we want to avoid the large dependency.
"""

import os
import platform
import enum
import resource
import requests
import time
import logging
import socket
from pathlib import Path
from typing import Union, Tuple, Optional
from urllib.parse import urlparse

__all__ = [
    "get_ac_state",
    "ACState",
    "get_inotify_limits",
    "CPU_COUNT",
    "cpu_usage_percent",
    "check_connection",
    "SystemdNotifier",
]


CPU_COUNT = os.cpu_count() or 1  # os.cpu_count can return None
LINUX_POWER_SUPPLY_PATH = "/sys/class/power_supply"


def multi_cat(*paths: Path) -> Union[int, bytes, None]:
    """
    Attempts to read the content of multiple files which may not exist. Returns the
    content of the first file which can be read. If none of them can be read return
    None. Returns an integer if the content is a digit.
    """
    for path in paths:
        try:
            ret = path.read_bytes().strip()
        except OSError:
            pass
        else:
            return int(ret) if ret.isdigit() else ret

    return None


class ACState(enum.Enum):
    """Enumeration of AC power states"""

    Connected = "Connected"
    Disconnected = "Disconnected"
    Undetermined = "Undetermined"


def get_ac_state() -> ACState:
    """
    Checks if the current device has AC power or is running on battery.

    :returns: ``True`` if the device has AC power, ``False`` otherwise.
    """
    if platform.system() == "Darwin":

        from ctypes import c_double
        from rubicon.objc.runtime import load_library

        iokit = load_library("IOKit")
        kIOPSTimeRemainingUnlimited = -2.0

        iokit.IOPSGetTimeRemainingEstimate.restype = c_double

        remaining_time = iokit.IOPSGetTimeRemainingEstimate()

        if remaining_time == kIOPSTimeRemainingUnlimited:
            return ACState.Connected
        else:
            return ACState.Disconnected

    elif platform.system() == "Linux":
        # taken from https://github.com/giampaolo/psutil

        supply_entry = list(os.scandir(LINUX_POWER_SUPPLY_PATH))

        ac_paths = [
            Path(entry.path)
            for entry in supply_entry
            if entry.name.startswith("A") or "ac" in entry.name.lower()
        ]

        battery_paths = [
            Path(entry.path)
            for entry in supply_entry
            if entry.name.startswith("B") or "battery" in entry.name.lower()
        ]

        online = multi_cat(*iter(path / "online" for path in ac_paths))

        if online is not None:
            if online == 1:
                return ACState.Connected
            else:
                return ACState.Disconnected

        elif len(battery_paths) > 0:
            # Get the first available battery. Usually this is "BAT0", except
            # some rare exceptions:
            # https://github.com/giampaolo/psutil/issues/1238
            bat0 = sorted(battery_paths)[0]

            try:
                status = (bat0 / "status").read_text().strip().lower()
            except OSError:
                status = ""

            if status == "discharging":
                return ACState.Disconnected
            elif status in ("charging", "full"):
                return ACState.Connected

    return ACState.Undetermined


def get_inotify_limits() -> Tuple[int, int, int]:
    """
    Returns the current inotify limit settings as tuple.

    :returns: ``(max_user_watches, max_user_instances, max_queued_events)``
    :raises OSError: if the settings cannot be read from /proc/sys/fs/inotify. This may
        happen if /proc/sys is left out of the kernel image or simply not mounted.
    """
    root = Path("/proc/sys/fs/inotify")

    max_user_watches_path = root / "max_user_watches"
    max_user_instances_path = root / "max_user_instances"
    max_queued_events_path = root / "max_queued_events"

    max_user_watches = int(max_user_watches_path.read_bytes().strip())
    max_user_instances = int(max_user_instances_path.read_bytes().strip())
    max_queued_events = int(max_queued_events_path.read_bytes().strip())

    return max_user_watches, max_user_instances, max_queued_events


def cpu_usage_percent(interval: float = 0.1) -> float:
    """
    Returns a float representing the CPU utilization of the current process as a
    percentage. This duplicates the similar method from psutil to avoid the psutil
    dependency.

    Compares process times to system CPU times elapsed before and after the interval
    (blocking). It is recommended for accuracy that this function be called with an
    interval of at least 0.1 sec.

    A value > 100.0 can be returned in case of processes running multiple threads on
    different CPU cores. The returned value is explicitly NOT split evenly between all
    available logical CPUs. This means that a busy loop process running on a system with
    2 logical CPUs will be reported as having 100% CPU utilization instead of 50%.

    :param interval: Interval in sec between comparisons of CPU times.
    :returns: CPU usage during interval in percent.
    """
    if interval <= 0:
        raise ValueError(f"interval is not positive (got {interval!r})")

    def timer() -> float:
        return time.monotonic() * CPU_COUNT

    st1 = timer()
    rt1 = resource.getrusage(resource.RUSAGE_SELF)
    time.sleep(interval)
    st2 = timer()
    rt2 = resource.getrusage(resource.RUSAGE_SELF)

    delta_proc = (rt2.ru_utime - rt1.ru_utime) + (rt2.ru_stime - rt1.ru_stime)
    delta_time = st2 - st1

    try:
        overall_cpus_percent = (delta_proc / delta_time) * 100
    except ZeroDivisionError:
        return 0.0
    else:
        single_cpu_percent = overall_cpus_percent * CPU_COUNT
        return round(single_cpu_percent, 1)


def check_connection(
    hostname: str, timeout: int = 2, logger: Optional[logging.Logger] = None
) -> bool:
    """
    A low latency check for an internet connection.

    :param hostname: Hostname to use for connection check.
    :param timeout: Timeout in seconds for connection check.
    :param logger: If provided, log output for connection failures will be logged to
        this logger with the level DEBUG.
    :returns: Connection availability.
    """
    if urlparse(hostname).scheme not in ["http", "https"]:
        hostname = "http://" + hostname
    try:
        requests.head(hostname, timeout=timeout)
        return True
    except Exception:
        if logger:
            logger.debug("Could not reach %s", hostname, exc_info=True)
        return False


class SystemdNotifier:
    """
    An interface to notify the systemd the service manager about status changes.

    Sends a status message to the systemd the service manager on the socket address
    provided by the NOTIFY_SOCKET environment variable. Does nothing NOTIFY_SOCKET is
    not set.

    See https://www.freedesktop.org/software/systemd/man/sd_notify.html for a
    documentation of message formats expected by systemd.
    """

    def __init__(self) -> None:
        self._socket = None

        addr = os.getenv("NOTIFY_SOCKET")
        if addr is None:
            return
        elif addr[0] == "@":
            addr = "\0" + addr[1:]

        try:
            self._socket = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
            self._socket.connect(addr)
        except OSError:
            pass

    def notify(self, status: str) -> None:
        """
        Send a status update to the service manager.

        :param status: The status update to send.
        """
        if self._socket:
            try:
                self._socket.sendall(status.encode(errors="replace"))
            except OSError:
                pass



"""
Dataclasses for our internal and external APIs.
"""

from __future__ import annotations

from enum import Enum
from dataclasses import dataclass
from datetime import datetime


# ==== user ============================================================================


class AccountType(Enum):
    """Enum of account types"""

    Basic = "basic"
    Pro = "pro"
    Business = "business"
    Other = "other"


@dataclass
class Team:
    id: str
    name: str


@dataclass
class RootInfo:
    root_namespace_id: str
    home_namespace_id: str


@dataclass
class UserRootInfo(RootInfo):
    pass


@dataclass
class TeamRootInfo(RootInfo):
    home_path: str


@dataclass
class Account:
    account_id: str
    display_name: str
    email: str
    email_verified: bool
    profile_photo_url: str | None
    disabled: bool


@dataclass
class FullAccount(Account):
    country: str | None
    locale: str
    team: Team | None
    team_member_id: str | None
    account_type: AccountType
    root_info: RootInfo


@dataclass
class TeamSpaceUsage:
    used: int
    allocation: int


@dataclass
class SpaceUsage:
    used: int
    allocated: int
    team_usage: TeamSpaceUsage | None


# ==== files ===========================================================================


class WriteMode(Enum):
    """Enum of write modes when uploading a file"""

    Add = "add"
    Update = "update"
    Overwrite = "overwrite"


@dataclass
class SharingInfo:
    read_only: bool


@dataclass
class Metadata:
    name: str
    path_lower: str
    path_display: str


@dataclass
class DeletedMetadata(Metadata):
    pass


@dataclass
class FileMetadata(Metadata):
    id: str
    client_modified: datetime
    server_modified: datetime
    rev: str
    size: int
    symlink_target: str | None
    shared: bool
    modified_by: str | None
    is_downloadable: bool
    content_hash: str


@dataclass
class FolderMetadata(Metadata):
    id: str
    shared: bool


@dataclass
class ListFolderResult:
    entries: list[Metadata]
    has_more: bool
    cursor: str


# ==== sharing =========================================================================


class LinkAccessLevel(Enum):
    """Enum of access levels to shared links"""

    Viewer = "viewer"
    Editor = "editor"
    Other = "other"


class LinkAudience(Enum):
    """Enum of shared link audience"""

    Public = "public"
    Team = "team"
    NoOne = "no_one"
    Other = "other"


@dataclass
class LinkPermissions:
    can_revoke: bool
    allow_download: bool
    effective_audience: LinkAudience
    link_access_level: LinkAccessLevel
    require_password: bool | None


@dataclass
class SharedLinkMetadata:
    url: str
    name: str
    path_lower: str | None
    expires: datetime | None
    link_permissions: LinkPermissions


@dataclass
class ListSharedLinkResult:
    entries: list[SharedLinkMetadata]
    has_more: bool
    cursor: str


# ==== update checks ===================================================================


@dataclass
class UpdateCheckResult:
    update_available: bool
    latest_release: str
    release_notes: str



from __future__ import annotations

import threading
from datetime import datetime
from os import path as osp
from typing import TYPE_CHECKING

import click

from rich.console import Console, ConsoleRenderable

from .dialogs import select_path, select, confirm, prompt, select_multiple
from .output import warn, ok, info, echo, RichDateField, rich_table
from .common import (
    convert_api_errors,
    check_for_fatal_errors,
    config_option,
    existing_config_option,
    inject_proxy,
)
from .core import DropboxPath, CliException
from ..core import FolderMetadata, SharedLinkMetadata
from ..utils.path import delete

if TYPE_CHECKING:
    from ..daemon import MaestralProxy
    from ..main import Maestral


OK = click.style("[OK]", fg="green")
FAILED = click.style("[FAILED]", fg="red")
KILLED = click.style("[KILLED]", fg="red")


def stop_daemon_with_cli_feedback(config_name: str) -> None:
    """Wrapper around :meth:`daemon.stop_maestral_daemon_process`
    with command line feedback."""

    from ..daemon import stop_maestral_daemon_process, Stop

    echo("Stopping Maestral...", nl=False)
    res = stop_maestral_daemon_process(config_name)
    if res == Stop.Ok:
        echo("\rStopping Maestral...        " + OK)
    elif res == Stop.NotRunning:
        echo("\rMaestral daemon is not running.")
    elif res == Stop.Killed:
        echo("\rStopping Maestral...        " + KILLED)
    elif res == Stop.Failed:
        echo("\rStopping Maestral...        " + FAILED)


def select_dbx_path_dialog(
    config_name: str, default_dir_name: str | None = None, allow_merge: bool = False
) -> str:
    """
    A CLI dialog to ask for a local Dropbox folder location.

    :param config_name: The configuration to use for the default folder name.
    :param default_dir_name: The default directory name. Defaults to
        "Dropbox ({config_name})" if not given.
    :param allow_merge: If ``True``, allows the selection of an existing folder without
        deleting it. Defaults to ``False``.
    :returns: Path given by user.
    """
    default_dir_name = default_dir_name or f"Dropbox ({config_name.capitalize()})"

    while True:
        res = select_path(
            "Please choose a local Dropbox folder:",
            default=f"~/{default_dir_name}",
            files_allowed=False,
        )
        res = res.rstrip(osp.sep)

        dropbox_path = osp.expanduser(res)

        if osp.exists(dropbox_path):
            if allow_merge:
                text = (
                    "Directory already exists. Do you want to replace it "
                    "or merge its content with your Dropbox?"
                )
                choice = select(text, options=["replace", "merge", "cancel"])
            else:
                text = (
                    "Directory already exists. Do you want to replace it? "
                    "Its content will be lost!"
                )
                replace = confirm(text)
                choice = 0 if replace else 2

            if choice == 0:
                err = delete(dropbox_path)
                if err:
                    warn(
                        "Could not write to selected location. "
                        "Please make sure that you have sufficient permissions."
                    )
                else:
                    ok("Replaced existing folder")
                    return dropbox_path
            elif choice == 1:
                ok("Merging with existing folder")
                return dropbox_path

        else:
            return dropbox_path


def link_dialog(m: MaestralProxy | Maestral) -> None:
    """
    A CLI dialog for linking a Dropbox account.

    :param m: Proxy to Maestral daemon.
    """
    authorize_url = m.get_auth_url()

    info(f"Linking new account for '{m.config_name}' config")
    info("Retrieving auth code from Dropbox")
    choice = select(
        "How would you like to you link your account?",
        options=["Open Dropbox website", "Print auth URL to console"],
    )

    if choice == 0:
        click.launch(authorize_url)
    else:
        info("Open the URL below to retrieve an auth code:")
        info(authorize_url)

    res = -1
    while res != 0:
        auth_code = prompt("Enter the auth code:")
        auth_code = auth_code.strip()

        res = m.link(auth_code)

        if res == 0:
            email = m.get_state("account", "email")
            ok(f"Linked to {email}")
        elif res == 1:
            warn("Invalid token, please try again")
        elif res == 2:
            warn("Could not connect to Dropbox, please try again")


@click.command(help="Start the sync daemon.")
@click.option(
    "--foreground",
    "-f",
    is_flag=True,
    default=False,
    help="Start Maestral in the foreground.",
)
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    default=False,
    help="Print log messages to stderr.",
)
@config_option
@convert_api_errors
def start(foreground: bool, verbose: bool, config_name: str) -> None:

    from ..daemon import (
        MaestralProxy,
        start_maestral_daemon,
        start_maestral_daemon_process,
        wait_for_startup,
        is_running,
        Start,
        CommunicationError,
    )

    if is_running(config_name):
        echo("Daemon is already running.")
        return

    @convert_api_errors
    def startup_dialog() -> None:

        try:
            wait_for_startup(config_name)
        except CommunicationError:
            return

        m = MaestralProxy(config_name)

        if m.pending_link:
            link_dialog(m)

        if m.pending_dropbox_folder:
            path = select_dbx_path_dialog(config_name, allow_merge=True)

            while True:
                try:
                    m.create_dropbox_directory(path)
                    break
                except OSError:
                    warn(
                        "Could not create folder. Please make sure that you have "
                        "permissions to write to the selected location or choose a "
                        "different location."
                    )

            include_all = confirm("Would you like sync all folders?")

            if not include_all:
                # get all top-level Dropbox folders
                info("Loading...")
                entries = m.list_folder("/", recursive=False)

                names = [e.name for e in entries if isinstance(e, FolderMetadata)]

                choices = select_multiple(
                    "Choose which folders to include", options=names
                )

                excluded_paths = [
                    f"/{name}"
                    for index, name in enumerate(names)
                    if index not in choices
                ]

                m.excluded_items = excluded_paths

            ok("Setup completed. Starting sync.")

        m.start_sync()

    if foreground:

        setup_thread = threading.Thread(target=startup_dialog, daemon=True)
        setup_thread.start()

        start_maestral_daemon(config_name, log_to_stderr=verbose)

    else:
        echo("Starting Maestral...", nl=False)

        res = start_maestral_daemon_process(config_name)

        if res == Start.Ok:
            echo("\rStarting Maestral...        " + OK)
        elif res == Start.AlreadyRunning:
            echo("\rStarting Maestral...        " + "Already running.")
        else:
            echo("\rStarting Maestral...        " + FAILED)
            echo("Please check logs for more information.")

        startup_dialog()


@click.command(help="Stop the sync daemon.")
@existing_config_option
def stop(config_name: str) -> None:
    stop_daemon_with_cli_feedback(config_name)


@click.command(help="Run the GUI if installed.")
@config_option
def gui(config_name: str) -> None:

    from packaging.version import Version
    from packaging.requirements import Requirement
    from importlib_metadata import entry_points, requires, version

    # Find all entry points for "maestral_gui" registered by other packages.
    gui_entry_points = entry_points(group="maestral_gui")

    if len(gui_entry_points) == 0:
        raise CliException(
            "No maestral GUI installed. Please run 'pip3 install maestral[gui]'."
        )

    entry_point_names = [e.name for e in gui_entry_points]

    if len(entry_point_names) > 1:
        index = select("Multiple GUIs found, please choose:", entry_point_names)
    else:
        index = 0

    entry_point = gui_entry_points[entry_point_names[index]]

    if entry_point in {"maestral_cocoa", "maestral_qt"}:
        # For 1st party GUIs "maestral_cocoa" or "maestral_qt", check if the installed
        # version fulfills requirements in maestral's gui extra.
        requirement_names = requires("maestral")
        if requirement_names is not None:
            for name in requirement_names:
                r = Requirement(name)
                if r.marker and r.marker.evaluate({"extra": "gui"}):
                    version_str = version(r.name)
                    if not r.specifier.contains(Version(version_str), prereleases=True):
                        raise CliException(
                            f"{r.name}{r.specifier} required but you have {version_str}"
                        )

    # Run the GUI.
    run = entry_point.load()
    run(config_name)


@click.command(help="Pause syncing.")
@inject_proxy(fallback=False, existing_config=True)
def pause(m: Maestral) -> None:
    m.stop_sync()
    ok("Syncing paused.")


@click.command(help="Resume syncing.")
@inject_proxy(fallback=False, existing_config=True)
def resume(m: Maestral) -> None:
    if not check_for_fatal_errors(m):
        m.start_sync()
        ok("Syncing resumed.")


@click.group(help="Link, unlink and view the Dropbox account.")
def auth() -> None:
    pass


@auth.command(name="link", help="Link a new Dropbox account.")
@click.option(
    "--relink",
    "-r",
    is_flag=True,
    default=False,
    help="Relink to the existing account. Keeps the sync state.",
)
@click.option(
    "--refresh-token",
    hidden=True,
    help="Refresh token to bypass OAuth exchange.",
)
@click.option(
    "--access-token",
    hidden=True,
    help="Access token to bypass OAuth exchange.",
)
@inject_proxy(fallback=True, existing_config=False)
@convert_api_errors
def auth_link(
    m: Maestral, relink: bool, refresh_token: str | None, access_token: str | None
) -> None:
    if m.pending_link or relink:
        if refresh_token or access_token:
            m.link(refresh_token=refresh_token, access_token=access_token)
        else:
            link_dialog(m)
    else:
        echo(
            "Maestral is already linked. Use '-r' to relink to the same "
            "account or specify a new config name with '-c'."
        )


@auth.command(
    name="unlink",
    help="""
Unlink your Dropbox account.

If Maestral is running, it will be stopped before unlinking.
""",
)
@click.option(
    "--yes", "-Y", is_flag=True, default=False, help="Skip confirmation prompt."
)
@existing_config_option
@convert_api_errors
def auth_unlink(yes: bool, config_name: str) -> None:
    if not yes:
        yes = confirm("Are you sure you want unlink your account?", default=False)

    if yes:
        from ..main import Maestral

        stop_daemon_with_cli_feedback(config_name)
        m = Maestral(config_name)
        m.unlink()

        ok("Unlinked Maestral.")


@auth.command(name="status", help="View authentication status.")
@existing_config_option
def auth_status(config_name: str) -> None:

    from ..config import MaestralConfig, MaestralState

    conf = MaestralConfig(config_name)
    state = MaestralState(config_name)

    dbid = conf.get("auth", "account_id")
    email = state.get("account", "email")
    account_type = state.get("account", "type").capitalize()

    echo("")
    echo(f"Email:         {email}")
    echo(f"Account type:  {account_type}")
    echo(f"Dropbox ID:    {dbid}")
    echo("")


@click.group(help="Create and manage shared links.")
def sharelink() -> None:
    pass


@sharelink.command(
    name="create", help="Create a shared link for a file or folder. Return the URL."
)
@click.argument("dropbox_path", type=DropboxPath())
@click.option(
    "-p",
    "--password",
    help="Optional password for the link.",
)
@click.option(
    "-e",
    "--expiry",
    metavar="DATE",
    type=click.DateTime(formats=["%Y-%m-%d", "%Y-%m-%dT%H:%M", "%Y-%m-%d %H:%M"]),
    help="Expiry time for the link (e.g. '2025-07-24 20:50').",
)
@inject_proxy(fallback=True, existing_config=True)
@convert_api_errors
def sharelink_create(
    m: Maestral,
    dropbox_path: str,
    password: str,
    expiry: datetime | None,
) -> None:
    link_info = m.create_shared_link(dropbox_path, password=password, expires=expiry)
    echo(link_info.url)


@sharelink.command(name="revoke", help="Revoke a shared link.")
@click.argument("url")
@inject_proxy(fallback=True, existing_config=True)
@convert_api_errors
def sharelink_revoke(m: Maestral, url: str) -> None:
    m.revoke_shared_link(url)
    ok("Revoked shared link.")


@sharelink.command(
    name="list", help="List shared links for given paths or all shared links."
)
@click.argument("dropbox_path", nargs=-1, type=DropboxPath())
@click.option(
    "-l",
    "--long",
    is_flag=True,
    default=False,
    help="Show output in long format with metadata.",
)
@inject_proxy(fallback=True, existing_config=True)
@convert_api_errors
def sharelink_list(m: Maestral, dropbox_path: list[str], long: bool) -> None:
    links: list[SharedLinkMetadata]

    if len(dropbox_path) > 0:
        links = []
        for dbx_path in dropbox_path:
            links.extend(m.list_shared_links(dbx_path))
    else:
        links = m.list_shared_links()

    if long:
        link_table = rich_table("URL", "Item", "Access", "Expires")

        for link in links:
            dt_field: ConsoleRenderable | str

            if link.expires:
                dt_field = RichDateField(link.expires)
            else:
                dt_field = "-"

            if link.link_permissions.require_password:
                access = "password"
            else:
                access = link.link_permissions.effective_audience.value

            link_table.add_row(link.url, link.name, access, dt_field)

        console = Console()
        console.print(link_table)

    else:
        echo("\n".join(link.url for link in links))



"""This module defines custom logging records and handlers."""

from __future__ import annotations

import os
import concurrent.futures
import logging
from logging.handlers import RotatingFileHandler
from collections import deque
from concurrent.futures import Future
from typing import Type, Sequence

from .config import MaestralConfig
from .utils import sanitize_string
from .utils.appdirs import get_log_path
from .utils.integration import SystemdNotifier

InvalidStateError: Type[Exception]

try:
    from concurrent.futures import (  # type:ignore[attr-defined, no-redef]
        InvalidStateError,
    )
except ImportError:
    # Python 3.7 and lower
    InvalidStateError = RuntimeError


__all__ = [
    "CachedHandler",
    "SdNotificationHandler",
    "EncodingSafeLogRecord",
    "scoped_logger",
    "scoped_logger_name",
    "LOG_FMT_LONG",
    "LOG_FMT_SHORT",
    "setup_logging",
]

LOG_FMT_LONG = logging.Formatter(
    fmt="%(asctime)s %(module)s %(levelname)s: %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)
LOG_FMT_SHORT = logging.Formatter(fmt="%(message)s")


class EncodingSafeLogRecord(logging.LogRecord):
    """A log record which ensures that messages contain only unicode characters

    This is useful when log messages may contain file paths generates by OS APIs. In
    Python, such path strings may contain surrogate escapes and will therefore raise
    a :exc:`UnicodeEncodeError` under many circumstances (printing to stdout, etc.).
    """

    def getMessage(self) -> str:
        """
        Formats the log message and replaces all surrogate escapes with "�".
        """
        msg = super().getMessage()
        return sanitize_string(msg)


logging.setLogRecordFactory(EncodingSafeLogRecord)


class CachedHandler(logging.Handler):
    """Handler which stores past records

    This is used to populate Maestral's status and error interfaces. The method
    :meth:`wait_for_emit` can be used from another thread to block until a new record is
    emitted, for instance to react to state changes.

    :param level: Initial log level. Defaults to NOTSET.
    :param maxlen: Maximum number of records to store. If ``None``, all records will be
        stored. Defaults to ``None``.
    """

    cached_records: deque[logging.LogRecord]
    _emit_future: Future[bool]

    def __init__(self, level: int = logging.NOTSET, maxlen: int | None = None) -> None:
        super().__init__(level=level)
        self.cached_records = deque([], maxlen)
        self._emit_future = Future()

    def emit(self, record: logging.LogRecord) -> None:
        """
        Logs the specified log record and saves it to the cache.

        :param record: Log record.
        """
        self.cached_records.append(record)

        # notify any waiting coroutines that we have a status change
        try:
            self._emit_future.set_result(True)
        except InvalidStateError:
            pass

    def wait_for_emit(self, timeout: float | None) -> bool:
        """
        Blocks until a new record is emitted. This is effectively a longpoll API.

        :param timeout: Maximum time to block before returning.
        :returns: ``True`` if there was a status change, ``False`` in case of a timeout.
        """
        try:
            self._emit_future.result(timeout=timeout)
        except concurrent.futures.TimeoutError:
            return False

        self._emit_future = Future()  # reset future
        return True

    def getLastMessage(self) -> str:
        """
        :returns: The log message of the last record or an empty string.
        """
        try:
            last_record = self.cached_records[-1]
            return last_record.getMessage()
        except IndexError:
            return ""

    def getAllMessages(self) -> list[str]:
        """
        :returns: A list of all record messages.
        """
        return [r.getMessage() for r in self.cached_records]

    def clear(self) -> None:
        """
        Clears all cached records.
        """
        self.cached_records.clear()


class SdNotificationHandler(logging.Handler):
    """Handler which emits messages as systemd notifications

    This is useful when used from a systemd service and will do nothing when no
    NOTIFY_SOCKET is provided.
    """

    notifier = SystemdNotifier()

    def emit(self, record: logging.LogRecord) -> None:
        """
        Sends the record message to systemd as service status.

        :param record: Log record.
        """
        self.notifier.notify(f"STATUS={record.getMessage()}")


def scoped_logger_name(module_name: str, config_name: str = "maestral") -> str:
    """
    Returns a logger name for the module ``module_name``, scoped to the given config.

    :param module_name: Module name.
    :param config_name: Config name.
    :returns: Scoped logger name.
    """
    if config_name == "maestral":
        return module_name
    else:
        return f"{config_name}-{module_name}"


def scoped_logger(module_name: str, config_name: str = "maestral") -> logging.Logger:
    """
    Returns a logger for the module ``module_name``, scoped to the given config.

    :param module_name: Module name.
    :param config_name: Config name.
    :returns: Logger instances scoped to the config.
    """
    return logging.getLogger(scoped_logger_name(module_name, config_name))


def setup_logging(
    config_name: str,
    file: bool = True,
    stderr: bool = True,
    journal: bool = True,
    status: bool = True,
) -> Sequence[logging.Handler]:
    """
    Set up loging to external channels. Systemd-related logging will fail silently if
    the current process was not started by systemd.

    :param config_name: Config name to determine log level and namespace for loggers.
        See :meth:`scoped_logger_name` for how the logger name is determined.
    :param file: Whether to log to files.
    :param stderr: Whether to log to stderr.
    :param journal: Whether to log to the systemd journal.
    :param status: Whether to log to the systemd status notifier. Note that this will
        always be performed at level INFO.
    :returns: Log handlers.
    """
    level = MaestralConfig(config_name).get("app", "log_level")
    root_logger = scoped_logger("maestral", config_name)
    root_logger.setLevel(min(level, logging.INFO))

    handlers: list[logging.Handler] = []

    # Log to file.
    if file:
        logfile = get_log_path("maestral", f"{config_name}.log")
        log_handler_file = RotatingFileHandler(logfile, maxBytes=10**7, backupCount=1)
        log_handler_file.setFormatter(LOG_FMT_LONG)
        log_handler_file.setLevel(level)
        root_logger.addHandler(log_handler_file)
        handlers.append(log_handler_file)

    # Log to systemd journal when launched as systemd service.
    if journal and os.getenv("INVOCATION_ID"):
        try:
            from systemd.journal import JournalHandler
        except ImportError:
            pass
        else:
            log_handler_journal = JournalHandler(SYSLOG_IDENTIFIER="maestral")
            log_handler_journal.setFormatter(LOG_FMT_SHORT)
            log_handler_journal.setLevel(level)
            root_logger.addHandler(log_handler_journal)
            handlers.append(log_handler_journal)

    # Log to systemd notify status when launched as systemd service.
    if status and os.getenv("NOTIFY_SOCKET"):
        log_handler_sd = SdNotificationHandler()
        log_handler_sd.setFormatter(LOG_FMT_SHORT)
        log_handler_sd.setLevel(logging.INFO)
        root_logger.addHandler(log_handler_sd)
        handlers.append(log_handler_sd)

    # Log to stderr if requested.
    if stderr:
        log_handler_stream = logging.StreamHandler()
        log_handler_stream.setFormatter(LOG_FMT_LONG)
        log_handler_stream.setLevel(level)
        root_logger.addHandler(log_handler_stream)
        handlers.append(log_handler_stream)

    return handlers



from __future__ import annotations

import os
import io
import ast
import logging
import textwrap
from os import path as osp
from typing import cast, TYPE_CHECKING

import click
from click.shell_completion import get_completion_class
from rich.console import Console
from rich.text import Text

from .cli_core import select_dbx_path_dialog
from .dialogs import confirm, select
from .output import ok, warn, echo, echo_via_pager, RichDateField, rich_table
from .utils import get_term_size
from .common import convert_api_errors, existing_config_option, inject_proxy
from .core import DropboxPath, ConfigKey, CliException

if TYPE_CHECKING:
    from ..main import Maestral


@click.command(help="Move the local Dropbox folder.")
@click.argument("new_path", required=False, type=click.Path(writable=True))
@inject_proxy(fallback=True, existing_config=True)
def move_dir(m: Maestral, new_path: str) -> None:
    new_path = new_path or select_dbx_path_dialog(m.config_name)
    new_path = osp.realpath(osp.expanduser(new_path))

    m.move_dropbox_directory(new_path)

    ok(f"Dropbox folder moved to {new_path}.")


@click.command(
    help="""
Rebuild the sync index.

Rebuilding may take several minutes, depending on the size of your Dropbox.
""",
)
@click.option(
    "--yes", "-Y", is_flag=True, default=False, help="Skip confirmation prompt."
)
@inject_proxy(fallback=True, existing_config=True)
@convert_api_errors
def rebuild_index(m: Maestral, yes: bool) -> None:
    size = get_term_size()

    msg = textwrap.fill(
        "Rebuilding the index may take several minutes, depending on the size of "
        "your Dropbox. Any changes to local files will be synced once rebuilding "
        "has completed. If you stop the daemon during the process, rebuilding will "
        "start again on the next launch.\nIf the daemon is not currently running, "
        "a rebuild will be scheduled for the next startup.",
        width=size.columns,
    )

    echo(msg + "\n")

    if yes or confirm("Do you want to continue?", default=False):

        m.rebuild_index()

        if m.running:
            ok("Rebuilding now. Run 'maestral status' to view progress.")
        else:
            ok("Sync is not running. Rebuilding scheduled for next startup.")


@click.command(help="List old file revisions.")
@click.argument("dropbox_path", type=DropboxPath())
@click.option(
    "-l",
    "--limit",
    help="Maximum number of revs to list.",
    show_default=True,
    type=click.IntRange(min=1, max=100),
    default=10,
)
@inject_proxy(fallback=True, existing_config=True)
@convert_api_errors
def revs(m: Maestral, dropbox_path: str, limit: int) -> None:
    table = rich_table("Revision", "Modified Time")

    for entry in m.list_revisions(dropbox_path, limit=limit):
        table.add_row(Text(entry.rev), RichDateField(entry.client_modified))

    console = Console()
    console.print(table)


@click.command(
    help="""
Compare two revisions of a file.

If no revs are passed to the command, you can select the revisions interactively. If
only one rev is passed, it is compared to the local version of the file. The diff is
shown via a pager if longer 30 lines.

Warning: The specified revisions will be downloaded to temp files and loaded into memory
to generate the diff. Depending on the file size, this may use significant disk space
and memory.
""",
)
@click.argument("dropbox_path", type=DropboxPath())
@click.option(
    "-v",
    "--rev",
    help="Revisions to compare (multiple allowed).",
    multiple=True,
    default=[],
)
@click.option("--no-color", help="Don't use colors for the diff.", is_flag=True)
@click.option("--no-pager", help="Don't use a pager for output.", is_flag=True)
@click.option(
    "-l",
    "--limit",
    help="Maximum number of revs to list.",
    show_default=True,
    type=click.IntRange(min=1, max=100),
    default=10,
)
@inject_proxy(fallback=True, existing_config=True)
@convert_api_errors
def diff(
    m: Maestral,
    dropbox_path: str,
    rev: list[str],
    no_color: bool,
    no_pager: bool,
    limit: int,
) -> None:
    class LocalDummyFile:
        rev = None

    # Ask for user input if revs are not provided as CLI arguments.
    if len(rev) == 0:
        entries = m.list_revisions(dropbox_path, limit=limit)
        modified_dates: list[str] = []

        for entry in entries:
            field = RichDateField(entry.client_modified)
            modified_dates.append(field.format(40))

        dbx_path = entries[0].path_display
        local_path = m.to_local_path(dbx_path)

        if osp.isfile(local_path):
            # prepend local version as an option
            modified_dates.insert(0, "local version")
            entries.insert(0, LocalDummyFile())  # type: ignore

        index_base = select(
            message="New revision:",
            options=modified_dates,
            hint="(↓ to see more)" if len(entries) > 6 else "",
        )

        if index_base == len(entries) - 1:
            warn("Oldest revision selected, unable to find anything to compare.")
            return

        comparable_dates = modified_dates[index_base + 1 :]
        index_new = select(
            message="Old revision:",
            options=comparable_dates,
            hint="(↓ to see more)" if len(comparable_dates) > 6 else "",
        )

        old_rev = entries[index_new + index_base + 1].rev
        new_rev = entries[index_base].rev
    elif len(rev) == 1:
        old_rev = rev[0]
        new_rev = None
    elif len(rev) == 2:
        old_rev = rev[0]
        new_rev = rev[1]
    else:
        warn("You can only compare two revisions at a time.")
        return

    # Download up to two revisions to a local temporary folder
    # and compare them with a 'diff'. Only text files are supported.
    # If an unknown file type was found, everything that doesn't match
    # 'text/*', an error message gets printed.

    echo("Loading ...\r", nl=False)

    diff_output = m.get_file_diff(old_rev, new_rev)

    if len(diff_output) == 0:
        echo("There are no changes between the two revisions.")
        return

    def color(ind: int, line: str) -> str:
        """
        Color diff lines.
        Inspiration for colors was taken from the
        well known command 'git diff'.
        """

        if ind < 2:
            line = click.style(line, bold=True)
        elif line.startswith("+"):
            line = click.style(line, fg="green")
        elif line.startswith("-"):
            line = click.style(line, fg="red")
        # Don't highlight these in the intro.
        elif line.startswith("@@ "):
            line = click.style(line, fg="cyan")
        return line

    # Color the lines.
    if not no_color:
        diff_output = [color(i, l) for i, l in enumerate(diff_output)]

    # Enter pager if diff is too long
    if len(diff_output) > 30 and not no_pager:
        echo_via_pager("".join(diff_output))
    else:
        echo("".join(diff_output))


@click.command(
    help="""
Restore a previous version of a file.

If no revision number is given, old revisions will be listed.
""",
)
@click.argument("dropbox_path", type=DropboxPath())
@click.option("-v", "--rev", help="Revision to restore.", default="")
@click.option(
    "-l",
    "--limit",
    help="Maximum number of revs to list.",
    show_default=True,
    type=click.IntRange(min=1, max=100),
    default=10,
)
@inject_proxy(fallback=True, existing_config=True)
@convert_api_errors
def restore(m: Maestral, dropbox_path: str, rev: str, limit: int) -> None:
    if not rev:
        echo("Loading...\r", nl=False)
        entries = m.list_revisions(dropbox_path, limit=limit)
        dates = []
        for entry in entries:
            field = RichDateField(entry.client_modified)
            dates.append(field.format(40))

        index = select(
            message="Select a version to restore:",
            options=dates,
            hint="(↓ to see more)" if len(entries) > 6 else "",
        )

        rev = entries[index].rev

    m.restore(dropbox_path, rev)

    ok(f'Restored {rev} to "{dropbox_path}"')


@click.group(help="View and manage the log.")
def log() -> None:
    pass


@log.command(name="show", help="View logs with a pager in the console.")
@click.option(
    "--external", "-e", is_flag=True, default=False, help="Open logs in a GUI."
)
@existing_config_option
def log_show(external: bool, config_name: str) -> None:

    from ..utils.appdirs import get_log_path

    log_file = get_log_path("maestral", config_name + ".log")

    if external:
        res = click.launch(log_file)
    else:
        try:
            with open(log_file) as f:
                text = f.read()
            echo_via_pager(text)
        except OSError:
            res = 1
        else:
            res = 0

    if res > 0:
        raise CliException(f"Could not open log file at '{log_file}'")


@log.command(name="clear", help="Clear the log files.")
@existing_config_option
def log_clear(config_name: str) -> None:

    from ..utils.appdirs import get_log_path

    log_dir = get_log_path("maestral")
    log_name = config_name + ".log"

    log_files = []

    for file_name in os.listdir(log_dir):
        if file_name.startswith(log_name):
            log_files.append(os.path.join(log_dir, file_name))

    try:
        for file in log_files:
            open(file, "w").close()
        ok("Cleared log files.")
    except FileNotFoundError:
        ok("Cleared log files.")
    except OSError:
        raise CliException(
            f"Could not clear log at '{log_dir}'. " f"Please try to delete it manually"
        )


@log.command(name="level", help="Get or set the log level.")
@click.argument(
    "level_name",
    required=False,
    type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR"], case_sensitive=False),
)
@inject_proxy(fallback=True, existing_config=True)
def log_level(m: Maestral, level_name: str) -> None:
    if level_name:
        m.log_level = cast(int, getattr(logging, level_name))
        ok(f"Log level set to {level_name}.")
    else:
        level_name = logging.getLevelName(m.log_level)
        echo(f"Log level: {level_name}")


@click.group(
    help="""
Direct access to config values.

Warning: Changing some config values must be accompanied by maintenance tasks. For
example, changing the config value for the Dropbox location needs to be accompanied by
actually moving the folder. This command only gets / sets the value in the config file.
Most changes will also require a restart of the daemon to become effective.

Use the commands from the Settings section instead wherever possible. They will take
effect immediately, perform accompanying tasks for you, and never leave the daemon in an
inconsistent state.

Currently available config keys are:

\b
- path: the location of the local Dropbox folder
- excluded_items: list of files or folders excluded by selective sync
- account_id: the ID of the linked Dropbox account
- notification_level: the level for desktop notifications
- log_level: the log level.
- update_notification_interval: interval in secs to check for updates
- keyring: the keyring backend to use (full path of the class)
- reindex_interval: the interval in seconds for full reindexing
- max_cpu_percent: maximum CPU usage target per core
- keep_history: the sync history to keep in seconds
- upload: if upload sync is enabled
- download: if download sync is enabled
""",
)
def config() -> None:
    pass


@config.command(name="get", help="Print the value of a given configuration key.")
@click.argument("key", type=ConfigKey())
@inject_proxy(fallback=True, existing_config=True)
def config_get(m: Maestral, key: str) -> None:

    from ..config.main import KEY_SECTION_MAP

    # Check if the config key exists in any section.
    section = KEY_SECTION_MAP.get(key, "")

    if not section:
        raise CliException(f"'{key}' is not a valid configuration key.")

    value = m.get_conf(section, key)

    echo(value)


@config.command(
    name="set",
    help="""
Update configuration with a value for the given key.

Values will be cast to the proper type, raising an error where this is not possibly. For
instance, setting a boolean config value to 1 will actually set it to True.
""",
)
@click.argument("key", type=ConfigKey())
@click.argument("value")
@inject_proxy(fallback=True, existing_config=True)
@convert_api_errors
def config_set(m: Maestral, key: str, value: str) -> None:

    from ..config.main import KEY_SECTION_MAP, DEFAULTS_CONFIG

    section = KEY_SECTION_MAP.get(key, "")

    if not section:
        raise CliException(f"'{key}' is not a valid configuration key.")

    default_value = DEFAULTS_CONFIG[section][key]

    if isinstance(default_value, str):
        py_value = value
    else:
        try:
            py_value = ast.literal_eval(value)
        except (SyntaxError, ValueError):
            py_value = value

    try:
        m.set_conf(section, key, py_value)
    except ValueError as e:
        warn(e.args[0])


@config.command(name="show", help="Show all config keys and values")
@click.option("--no-pager", help="Don't use a pager for output.", is_flag=True)
@existing_config_option
def config_show(no_pager: bool, config_name: str) -> None:
    from ..config import MaestralConfig

    conf = MaestralConfig(config_name)

    with io.StringIO() as fp:
        conf.write(fp)
        if no_pager:
            echo(fp.getvalue())
        else:
            echo_via_pager(fp.getvalue())


@click.command(
    help="""
Generate completion script for your shell.

This command can generate shell completion scripts for bash, zsh or fish. Follow the
instructions below for your shell to load the resulting script. The exact config file
locations might vary based on your system. Make sure to restart your
shell before testing whether completions are working.

### bash

You can enable shell completion for all users by generating and saving the script as
follows:

\b
    maestral completion bash > /usr/share/bash-completion/completions/maestral

To enable shell completion for the current user only, save the script in a location of
your choice, for example `~/.local/completions/maestral`, and source it in `~/.bashrc`
by adding the line:

\b
    . ~/.local/completions/maestral

### zsh

Generate a `_maestral` completion script and put it somewhere in your `$fpath`. For
example:

\b
    maestral completion zsh > /usr/local/share/zsh/site-functions/_maestral

You can also save the completion script in a location of your choice and source it
in `~/.zshrc`. Ensure that the following is present in your `~/.zshrc`:

\b
    autoload -Uz compinit && compinit

### fish

Generate and save a `maestral.fish` completion script as follows. For all users:

\b
    maestral completion fish > /usr/share/fish/vendor_completions.d/maestral.fish

For the current user only:

\b
    maestral completion fish > ~/.config/fish/completions/maestral.fish

""",
)
@click.argument("shell", type=click.Choice(["bash", "zsh", "fish"]))
def completion(shell: str) -> None:

    from .cli_main import main

    comp_cls = get_completion_class(shell)

    if comp_cls is None:
        warn(f"{shell} shell is currently not supported")
        return

    comp = comp_cls(main, {}, "maestral", "_MAESTRAL_COMPLETE")

    try:
        echo(comp.source())
    except RuntimeError as exc:
        warn(exc.args[0])



"""
This module contains functions to retrieve platform dependent locations to store app
data. It supports macOS and Linux.
"""

# system imports
import os
import platform
from os import path as osp
from typing import Optional


__all__ = [
    "get_log_path",
    "get_cache_path",
    "get_autostart_path",
    "get_runtime_path",
    "get_conf_path",
    "get_home_dir",
    "get_data_path",
]


def to_full_path(
    path: str, subfolder: Optional[str], filename: Optional[str], create: bool
) -> str:
    if subfolder:
        path = osp.join(path, subfolder)
    if create:
        os.makedirs(path, exist_ok=True)
    if filename:
        path = osp.join(path, filename)
    return path


def get_home_dir() -> str:
    """
    Returns user home directory. This will be determined from the first
    valid result out of (osp.expanduser("~"), $HOME, $USERPROFILE, $TMP).
    """
    path = osp.expanduser("~")

    if osp.isdir(path):
        return path
    raise RuntimeError(
        "Please set the environment variable HOME to your user/home directory."
    )


home_dir = get_home_dir()


def get_conf_path(
    subfolder: Optional[str] = None, filename: Optional[str] = None, create: bool = True
) -> str:
    """
    Returns the default config path for the platform. This will be:

        - macOS: "~/Library/Application Support/<subfolder>/<filename>."
        - Linux: "XDG_CONFIG_HOME/<subfolder>/<filename>"
        - other: "~/.config/<subfolder>/<filename>"

    :param subfolder: The subfolder for the app.
    :param filename: The filename to append for the app.
    :param create: If ``True``, the folder ``subfolder`` will be created on-demand.
    """
    if platform.system() == "Darwin":
        conf_path = osp.join(get_home_dir(), "Library", "Application Support")
    elif platform.system() == "Linux":
        fallback = osp.join(get_home_dir(), ".config")
        conf_path = os.environ.get("XDG_CONFIG_HOME", fallback)
    else:
        raise RuntimeError("Platform not supported")

    return to_full_path(conf_path, subfolder, filename, create)


def get_data_path(
    subfolder: Optional[str] = None, filename: Optional[str] = None, create: bool = True
) -> str:
    """
    Returns the default path to save application data for the platform. This will be:

        - macOS: "~/Library/Application Support/SUBFOLDER/FILENAME"
        - Linux: "$XDG_DATA_DIR/SUBFOLDER/FILENAME"
        - fallback: "$HOME/.local/share/SUBFOLDER/FILENAME"

    Note: We do not use "~/Library/Saved Application State" on macOS since this folder
    is reserved for user interface state and can be cleared by the user / system.

    :param subfolder: The subfolder for the app.
    :param filename: The filename to append for the app.
    :param create: If ``True``, the folder ``subfolder`` will be created on-demand.
    """
    if platform.system() == "Darwin":
        state_path = osp.join(get_home_dir(), "Library", "Application Support")
    elif platform.system() == "Linux":
        fallback = osp.join(get_home_dir(), ".local", "share")
        state_path = os.environ.get("XDG_DATA_HOME", fallback)
    else:
        raise RuntimeError("Platform not supported")

    return to_full_path(state_path, subfolder, filename, create)


def get_cache_path(
    subfolder: Optional[str] = None, filename: Optional[str] = None, create: bool = True
) -> str:
    """
    Returns the default cache path for the platform. This will be:

        - macOS: "~/Library/Caches/SUBFOLDER/FILENAME"
        - Linux: "$XDG_CACHE_HOME/SUBFOLDER/FILENAME"
        - fallback: "$HOME/.cache/SUBFOLDER/FILENAME"

    :param subfolder: The subfolder for the app.
    :param filename: The filename to append for the app.
    :param create: If ``True``, the folder ``subfolder`` will be created on-demand.
    """
    if platform.system() == "Darwin":
        cache_path = osp.join(home_dir, "Library", "Caches")
    elif platform.system() == "Linux":
        fallback = osp.join(home_dir, ".cache")
        cache_path = os.environ.get("XDG_CACHE_HOME", fallback)
    else:
        raise RuntimeError("Platform not supported")

    return to_full_path(cache_path, subfolder, filename, create)


def get_log_path(
    subfolder: Optional[str] = None, filename: Optional[str] = None, create: bool = True
) -> str:
    """
    Returns the default log path for the platform. This will be:

        - macOS: "~/Library/Logs/SUBFOLDER/FILENAME"
        - Linux: "$XDG_CACHE_HOME/SUBFOLDER/FILENAME"
        - fallback: "$HOME/.cache/SUBFOLDER/FILENAME"

    :param subfolder: The subfolder for the app.
    :param filename: The filename to append for the app.
    :param create: If ``True``, the folder ``subfolder`` will be created on-demand.
    """
    if platform.system() == "Darwin":
        log_path = osp.join(home_dir, "Library", "Logs")
    elif platform.system() == "Linux":
        log_path = get_cache_path(create=False)
    else:
        raise RuntimeError("Platform not supported")

    return to_full_path(log_path, subfolder, filename, create)


def get_autostart_path(filename: Optional[str] = None, create: bool = True) -> str:
    """
    Returns the default path for login items for the platform. This will be:

        - macOS: "~/Library/LaunchAgents/FILENAME"
        - Linux: "$XDG_CONFIG_HOME/autostart/FILENAME"
        - fallback: "$HOME/.config/autostart/FILENAME"

    :param filename: The filename to append for the app.
    :param create: If ``True``, the folder ``subfolder`` will be created on-demand.
    """
    if platform.system() == "Darwin":
        autostart_path = osp.join(home_dir, "Library", "LaunchAgents")
    elif platform.system() == "Linux":
        autostart_path = get_conf_path("autostart", create=create)
    else:
        raise RuntimeError("Platform not supported")

    if filename:
        autostart_path = osp.join(autostart_path, filename)

    return autostart_path


def get_runtime_path(
    subfolder: Optional[str] = None, filename: Optional[str] = None, create: bool = True
) -> str:
    """
    Returns the default runtime path for the platform. This will be:

        - macOS: "~/Library/Application Support/SUBFOLDER/FILENAME"
        - Linux: "$XDG_RUNTIME_DIR/SUBFOLDER/FILENAME"
        - fallback: "$HOME/.cache/SUBFOLDER/FILENAME"

    :param subfolder: The subfolder for the app.
    :param filename: The filename to append for the app.
    :param create: If ``True``, the folder ``subfolder`` will be created on-demand.
    """
    if platform.system() == "Darwin":
        runtime_path = get_conf_path(create=False)
    elif platform.system() == "Linux":
        fallback = get_cache_path(create=False)
        runtime_path = os.environ.get("XDG_RUNTIME_DIR", fallback)
    else:
        raise RuntimeError("Platform not supported")

    return to_full_path(runtime_path, subfolder, filename, create)



from __future__ import annotations

from typing import TYPE_CHECKING

import click

from .output import echo, ok
from .common import convert_api_errors, existing_config_option, inject_proxy
from .core import DropboxPath, CliException

if TYPE_CHECKING:
    from ..main import Maestral


@click.command(
    help="""
Automatically start the sync daemon on login.

A systemd or launchd service will be created to start a sync daemon for the given
configuration on user login.
""",
)
@click.option("--yes", "-Y", is_flag=True, default=False)
@click.option("--no", "-N", is_flag=True, default=False)
@existing_config_option
def autostart(yes: bool, no: bool, config_name: str) -> None:

    from ..autostart import AutoStart

    auto_start = AutoStart(config_name)

    if not auto_start.implementation:
        echo(
            "Autostart is currently not supported for your platform.\n"
            "Autostart requires systemd on Linux or launchd on macOS."
        )
        return

    if yes or no:
        if yes:
            auto_start.enable()
            ok("Enabled start on login.")
        else:
            auto_start.disable()
            ok("Disabled start on login.")
    else:
        if auto_start.enabled:
            echo("Autostart is enabled. Use -N to disable.")
        else:
            echo("Autostart is disabled. Use -Y to enable.")


@click.group(help="View and manage excluded folders.")
def excluded() -> None:
    pass


@excluded.command(name="list", help="List all excluded files and folders.")
@inject_proxy(fallback=True, existing_config=True)
def excluded_list(m: Maestral) -> None:
    excluded_items = m.excluded_items
    excluded_items.sort()

    if len(excluded_items) == 0:
        echo("No excluded files or folders.")
    else:
        for item in excluded_items:
            echo(item)


@excluded.command(
    name="add",
    help="Add a file or folder to the excluded list and re-sync.",
)
@click.argument("dropbox_path", type=DropboxPath())
@inject_proxy(fallback=True, existing_config=True)
@convert_api_errors
def excluded_add(m: Maestral, dropbox_path: str) -> None:
    if dropbox_path == "/":
        raise CliException("Cannot exclude the root directory.")

    m.exclude_item(dropbox_path)
    ok(f"Excluded '{dropbox_path}'.")


@excluded.command(
    name="remove",
    help="""
Remove a file or folder from the excluded list and re-sync.

It is safe to call this method with items which have already been included, they will
not be downloaded again. If the given path lies inside an excluded folder, the parent
folder will be included as well (but no other items inside it).
""",
)
@click.argument("dropbox_path", type=DropboxPath())
@inject_proxy(fallback=False, existing_config=True)
@convert_api_errors
def excluded_remove(m: Maestral, dropbox_path: str) -> None:
    if dropbox_path == "/":
        return echo("The root directory is always included")

    m.include_item(dropbox_path)
    ok(f"Included '{dropbox_path}'. Now downloading...")


@click.group(help="Manage desktop notifications.")
def notify() -> None:
    pass


@notify.command(
    name="level",
    help="Get or set the level for desktop notifications.",
)
@click.argument(
    "level_name",
    required=False,
    type=click.Choice(["ERROR", "SYNCISSUE", "FILECHANGE"], case_sensitive=False),
)
@inject_proxy(fallback=True, existing_config=True)
def notify_level(m: Maestral, level_name: str) -> None:

    from .. import notify as _notify

    if level_name:
        m.notification_level = _notify.level_name_to_number(level_name)
        ok(f"Notification level set to {level_name}.")
    else:
        level_name = _notify.level_number_to_name(m.notification_level)
        echo(f"Notification level: {level_name}.")


@notify.command(
    name="snooze",
    help="Snooze desktop notifications of file changes.",
)
@click.argument("minutes", type=click.IntRange(min=0))
@inject_proxy(fallback=True, existing_config=True)
def notify_snooze(m: Maestral, minutes: int) -> None:
    m.notification_snooze = minutes

    if minutes > 0:
        ok(f"Notifications snoozed for {minutes} min. Set snooze to 0 to reset.")
    else:
        ok("Notifications enabled.")



from __future__ import annotations

import functools
import sys
from typing import Callable, Any, TypeVar, TYPE_CHECKING
from typing_extensions import ParamSpec

import click

from .core import ConfigName
from .output import warn
from .utils import get_term_size

if TYPE_CHECKING:
    from ..daemon import MaestralProxy
    from ..main import Maestral


P = ParamSpec("P")
T = TypeVar("T")


def convert_api_errors(func: Callable[P, T]) -> Callable[P, T]:
    """
    Decorator that catches a MaestralApiError and prints a formatted error message to
    stdout before exiting. Calls ``sys.exit(1)`` after printing the error to stdout.
    """

    from ..exceptions import MaestralApiError

    @functools.wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        try:
            return func(*args, **kwargs)
        except MaestralApiError as exc:
            warn(f"{exc.title}. {exc.message}")
            sys.exit(1)

    return wrapper


def check_for_fatal_errors(m: MaestralProxy | Maestral) -> bool:
    """
    Checks the given Maestral instance for fatal errors such as revoked Dropbox access,
    deleted Dropbox folder etc. Prints a nice representation to the command line.

    :param m: Proxy to Maestral daemon or Maestral instance.
    :returns: True in case of fatal errors, False otherwise.
    """

    import textwrap

    maestral_err_list = m.fatal_errors

    if len(maestral_err_list) > 0:

        size = get_term_size()

        err = maestral_err_list[0]
        wrapped_msg = textwrap.fill(err.message, width=size.columns)

        click.echo("")
        click.secho(err.title, fg="red")
        click.secho(wrapped_msg, fg="red")
        click.echo("")

        return True
    else:
        return False


config_option = click.option(
    "-c",
    "--config-name",
    default="maestral",
    type=ConfigName(existing=False),
    is_eager=True,
    expose_value=True,
    help="Run command with the given configuration.",
)
existing_config_option = click.option(
    "-c",
    "--config-name",
    default="maestral",
    type=ConfigName(),
    is_eager=True,
    expose_value=True,
    help="Run command with the given configuration.",
)


def inject_proxy(
    fallback: bool, existing_config: bool
) -> Callable[[Callable[P, T]], Callable[P, Any]]:
    def decorator(f: Callable[P, T]) -> Callable[P, Any]:
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> Any:

            from ..daemon import MaestralProxy, CommunicationError

            ctx = click.get_current_context()

            config_name = ctx.params.pop("config_name", "maestral")
            kwargs.pop("config_name", None)

            try:
                proxy = ctx.with_resource(MaestralProxy(config_name, fallback=fallback))
            except CommunicationError:
                click.echo("Maestral daemon is not running.")
                ctx.exit(0)
            else:
                return ctx.invoke(f, proxy, *args, **kwargs)

        if existing_config:
            f = existing_config_option(f)
        else:
            f = config_option(f)

        return functools.update_wrapper(wrapper, f)

    return decorator



from __future__ import annotations

import sys
import os
import time
from datetime import datetime
from typing import TYPE_CHECKING, Tuple

import click
from rich.console import Console, ConsoleRenderable
from rich.table import Column
from rich.text import Text
from rich.columns import Columns
from rich.filesize import decimal
from rich.progress import Progress, TextColumn, BarColumn, DownloadColumn, TaskID

from .output import echo, RichDateField, rich_table
from .common import convert_api_errors, check_for_fatal_errors, inject_proxy
from .core import DropboxPath
from ..models import SyncDirection, SyncEvent
from ..core import FileMetadata, FolderMetadata, DeletedMetadata

if TYPE_CHECKING:
    from ..main import Maestral


@click.command(help="Show the status of the daemon.")
@inject_proxy(fallback=False, existing_config=True)
@convert_api_errors
def status(m: Maestral) -> None:
    email = m.get_state("account", "email")
    account_type = m.get_state("account", "type").capitalize()
    usage = m.get_state("account", "usage")
    status_info = m.status

    account_str = f"{email} ({account_type})" if email else "--"
    usage_str = Text(usage or "--")

    n_errors = len(m.sync_errors)
    color = "red" if n_errors > 0 else "green"
    n_errors_str = Text(str(n_errors), style=color)

    status_table = rich_table()
    status_table.add_row("Account", account_str)
    status_table.add_row("Usage", usage_str)
    status_table.add_row("Status", status_info)
    status_table.add_row("Sync errors", n_errors_str)

    console = Console()

    console.print("")
    console.print(status_table, highlight=False)
    console.print("")

    check_for_fatal_errors(m)

    sync_errors = m.sync_errors

    if len(sync_errors) > 0:
        sync_errors_table = rich_table("Path", "Error")

        for error in sync_errors:
            sync_errors_table.add_row(error.dbx_path, f"{error.title}. {error.message}")

        console.print(sync_errors_table)
        console.print("")


@click.command(
    help="""
Show the sync status of a local file or folder.

Returned value will be 'uploading', 'downloading', 'up to date', 'error', or 'unwatched'
(for files outside of the Dropbox directory). This will always be 'unwatched' if syncing
is paused. This command can be used to for instance to query information for a plugin to
a file-manager.
""",
)
@click.argument("local_path", type=click.Path(exists=True, resolve_path=True))
@inject_proxy(fallback=True, existing_config=True)
@convert_api_errors
def filestatus(m: Maestral, local_path: str) -> None:
    stat = m.get_file_status(local_path)
    echo(stat)


@click.command(help="Live view of all items being synced.")
@inject_proxy(fallback=False, existing_config=True)
@convert_api_errors
def activity(m: Maestral) -> None:
    if check_for_fatal_errors(m):
        return

    from Pyro5.errors import ConnectionClosedError

    EventKey = Tuple[str, SyncDirection]
    progressbar_for_path: dict[EventKey, TaskID] = {}

    def _event_key(e: SyncEvent) -> EventKey:
        return e.dbx_path, e.direction

    console = Console()

    arrow = {SyncDirection.Up: "↑", SyncDirection.Down: "↓"}

    try:
        with console.screen():
            with Progress(
                TextColumn("[bold bright_blue]{task.description}"),
                TextColumn("[bright_blue]{task.fields[filename]}"),
                TextColumn(" "),
                BarColumn(bar_width=None),
                TextColumn("[progress.percentage]{task.percentage:>3.1f}%"),
                TextColumn("•"),
                DownloadColumn(),
                auto_refresh=False,
                console=console,
            ) as progress:
                while True:
                    msg = f"\rStatus: {m.status}, Sync errors: {len(m.sync_errors)}"
                    progress.console.clear()
                    progress.console.print(msg)

                    sync_events = m.get_activity(limit=console.height - 1)
                    event_keys = set(_event_key(e) for e in sync_events)

                    for key, task_id in progressbar_for_path.copy().items():
                        if key not in event_keys:
                            progress.remove_task(task_id)
                            progressbar_for_path.pop(key)

                    for event in sync_events:
                        try:
                            task_id = progressbar_for_path[_event_key(event)]
                        except KeyError:
                            task_id = progress.add_task(
                                f"{arrow[event.direction]} {event.change_type.name}",
                                total=event.size,
                                completed=event.completed,
                                filename=os.path.basename(event.dbx_path),
                            )
                            progressbar_for_path[_event_key(event)] = task_id
                        else:
                            progress.update(task_id, completed=event.completed)

                    time.sleep(0.2)
                    progress.refresh()
    except ConnectionClosedError:
        return echo("Maestral daemon is not running.")


@click.command(help="Show recently changed or added files.")
@inject_proxy(fallback=True, existing_config=True)
@convert_api_errors
def history(m: Maestral) -> None:
    events = m.get_history()
    table = rich_table("Path", "Change", "Location", "Time")

    for event in events:
        dt_local_naive = datetime.fromtimestamp(event.change_time_or_sync_time)
        location = "local" if event.direction is SyncDirection.Up else "remote"
        table.add_row(
            Text(event.dbx_path, overflow="ellipsis", no_wrap=True),
            Text(event.change_type.value),
            Text(location),
            RichDateField(dt_local_naive),
        )

    console = Console()
    console.print(table)


@click.command(help="List contents of a Dropbox directory.")
@click.argument("dropbox_path", type=DropboxPath(), default="")
@click.option(
    "-l",
    "--long",
    is_flag=True,
    default=False,
    help="Show output in long format with metadata.",
)
@click.option(
    "-d",
    "--include-deleted",
    is_flag=True,
    default=False,
    help="Include deleted items in listing.",
)
@inject_proxy(fallback=True, existing_config=True)
@convert_api_errors
def ls(m: Maestral, long: bool, dropbox_path: str, include_deleted: bool) -> None:
    echo("Loading...\r", nl=False)

    entries_iter = m.list_folder_iterator(
        dropbox_path,
        recursive=False,
        include_deleted=include_deleted,
    )

    entries = [entry for entries in entries_iter for entry in entries]
    entries.sort(key=lambda e: e.name)

    console = Console()

    if long:
        table = rich_table(
            Column("Name"),
            Column("Type"),
            Column("Size", justify="right"),
            Column("Shared"),
            Column("Syncing"),
            Column("Last Modified"),
        )

        for entry in entries:
            text = "shared" if getattr(entry, "shared", False) else "private"
            color = "bright_black" if text == "private" else ""
            shared_field = Text(text, style=color)

            excluded_status = m.excluded_status(entry.path_lower)
            color = "green" if excluded_status == "included" else ""
            text = "✓" if excluded_status == "included" else excluded_status
            excluded_field = Text(text, style=color)

            dt_field: ConsoleRenderable

            if isinstance(entry, FileMetadata):
                size = decimal(entry.size)
                dt_field = RichDateField(entry.client_modified)
                item_type = "file"
            elif isinstance(entry, FolderMetadata):
                size = "-"
                dt_field = Text("-")
                item_type = "folder"
            else:
                size = "-"
                dt_field = Text("-")
                item_type = "deleted"

            table.add_row(
                Text(entry.name, overflow="ellipsis", no_wrap=True),
                item_type,
                size,
                shared_field,
                excluded_field,
                dt_field,
            )

        console.print(table)

    elif not sys.stdout.isatty():
        names = [entry.name for entries in entries_iter for entry in entries]
        console.print("\n".join(names))

    else:
        fields: list[Text] = []

        for entry in entries:
            color = "blue" if isinstance(entry, DeletedMetadata) else ""
            fields.append(Text(entry.name, style=color))

        max_len = max(len(f) for f in fields)
        console.print(Columns(fields, width=max_len, column_first=True))


@click.command(help="List all configured Dropbox accounts.")
@click.option(
    "--clean",
    is_flag=True,
    default=False,
    help="Remove config files without a linked account.",
)
def config_files(clean: bool) -> None:

    from ..daemon import is_running
    from ..config import (
        MaestralConfig,
        MaestralState,
        list_configs,
        remove_configuration,
    )

    if clean:
        # Clean up stale config files.
        for name in list_configs():
            conf = MaestralConfig(name)
            dbid = conf.get("auth", "account_id")

            if dbid == "" and not is_running(name):
                remove_configuration(name)
                echo(f"Removed: {conf.config_path}")

    else:
        # Display config files.
        table = rich_table("Config name", "Account", "Path")
        for name in list_configs():
            conf = MaestralConfig(name)
            state = MaestralState(name)

            table.add_row(
                name,
                state.get("account", "email"),
                Text(conf.config_path, overflow="ellipsis", no_wrap=True),
            )

        console = Console()
        console.print(table)



"""
This model defines our core SQLite database interface.
"""

from __future__ import annotations

from typing import Any

import sqlite3


class Database:
    """Wrapper around sqlite3.Connection with atomic transactions."""

    def __init__(self, connection: sqlite3.Connection) -> None:
        connection.row_factory = sqlite3.Row
        self.connection = connection

    def close(self) -> None:
        """Closes the SQL connection."""
        self.connection.close()

    def execute(self, sql: str, *args: Any) -> sqlite3.Cursor:
        """
        Creates a cursor and executes the given SQL statement.

        :param sql: SQL statement to execute.
        :param args: Parameters to substitute for placeholders in SQL statement.
        :returns: The created cursor.
        """
        with self.connection:
            return self.connection.execute(sql, args)

    def executescript(self, script: str) -> None:
        """
        Creates a cursor and executes the given SQL script.

        :param script: SQL script to execute.
        :returns: The created cursor.
        """
        with self.connection:
            self.connection.cursor().executescript(script)



"""Utility modules and functions"""

from __future__ import annotations

import os
from types import TracebackType
from typing import Iterator, TypeVar, Optional, Iterable, Tuple, Type

from packaging.version import Version


_N = TypeVar("_N", float, int)
_T = TypeVar("_T")
_ExecInfoType = Tuple[Type[BaseException], BaseException, Optional[TracebackType]]


def natural_size(num: float, unit: str = "B", sep: bool = True) -> str:
    """
    Convert number to a human-readable string with decimal prefix.

    :param float num: Value in given unit.
    :param unit: Unit suffix.
    :param sep: Whether to separate unit and value with a space.
    :returns: Human readable string with decimal prefixes.
    """
    sep_char = " " if sep else ""

    for prefix in ("", "K", "M", "G"):
        if abs(num) < 1000.0:
            return f"{num:3.1f}{sep_char}{prefix}{unit}"
        num /= 1000.0

    prefix = "T"
    return f"{num:.1f}{sep_char}{prefix}{unit}"


def chunks(lst: list[_T], n: int, consume: bool = False) -> Iterator[list[_T]]:
    """
    Partitions a list into chunks of length ``n``.

    :param lst: Iterable to partition.
    :param n: Chunk size.
    :param consume: If True, the list will be consumed (emptied) during the iteration.
    :returns: Iterator over chunks.
    """
    if consume:
        while lst:
            chunk = lst[0:n]
            del lst[0:n]
            yield chunk
    else:
        for i in range(0, len(lst), n):
            yield lst[i : i + n]


def clamp(n: _N, minn: _N, maxn: _N) -> _N:
    """
    Clamps a number between a minimum and maximum value.

    :param n: Original value.
    :param minn: Minimum allowed value.
    :param maxn: Maximum allowed value.
    :returns: Clamped value.
    """
    if n > maxn:
        return maxn
    elif n < minn:
        return minn
    else:
        return n


def get_newer_version(version: str, releases: Iterable[str]) -> Optional[str]:
    """
    Checks a given release version against a version list of releases to see if an
    update is available. Only offers newer versions if they are not a prerelease.

    :param version: The current version.
    :param releases: A list of valid cleaned releases.
    :returns: The version string of the latest release if a newer release is available.
    """
    releases = [r for r in releases if not Version(r).is_prerelease]
    releases.sort(key=Version)
    latest_release = releases[-1]

    return latest_release if Version(version) < Version(latest_release) else None


def removeprefix(string: str, prefix: str) -> str:
    """
    Removes the given prefix from a string. Only the first instance of the prefix is
    removed. The original string is returned if it does not start with the given prefix.

    This follows the Python 3.9 implementation of ``str.removeprefix``.

    :param string: Original string.
    :param prefix: Prefix to remove.
    :returns: String without prefix.
    """
    if string.startswith(prefix):
        return string[len(prefix) :]
    else:
        return string[:]


def sanitize_string(string: str) -> str:
    """
    Converts a string provided by file system APIs, which may contain surrogate escapes
    for bytes with unknown encoding, to a string which can always be displayed or
    printed. This is done by replacing invalid characters with "�".

    :param string: Original string.
    :returns: Sanitised path where all surrogate escapes have been replaced with "�".
    """
    return os.fsencode(string).decode(errors="replace")


def exc_info_tuple(exc: BaseException) -> _ExecInfoType:
    """Creates an exc-info tuple from an exception."""
    return type(exc), exc, exc.__traceback__
